"""Utilities for working with collections."""

# Python Modules
from types import GeneratorType
from typing import Any, Iterable, Optional


# 3rd Party Modules

# Project Modules


def is_iterable(obj: Any, consider_string_iterable: bool = False) -> bool:
    """
    Tests if the object is iterable.

    Parameters
    ----------
    obj : any
        An object.
    consider_string_iterable : bool, default = False
        Whether to consider strings iterable or not.

    Returns
    -------
    bool
        ``True`` if ``obj`` is iterable.
    """
    if isinstance(obj, str):
        return consider_string_iterable

    try:
        iter(obj)
    except TypeError:
        return False
    else:
        return True


def is_generator(obj: Any) -> bool:
    """
    Checks if the object is a generator.

    Parameters
    ----------
    obj : any
        An object.

    Returns
    -------
    bool
        ``True`` if the object is a generator.

    """
    return isinstance(obj, GeneratorType)


def get_first_non_null_value(collection: Iterable) -> Optional[Any]:
    """Recursively try to get the first non-null value in a collection.

    This method will recursively traverse the collection until it finds
    a non-iterable value that is not ``None``.

    Parameters
    ----------
    collection : Iterable
        The collection to retrieve the value from.

    Returns
    -------
    Any, optional
        The first non-null value in the series, if one exists, otherwise return
        ``None``.

    """
    for value in collection:
        if is_iterable(value) and not isinstance(value, str):
            value = get_first_non_null_value(value)

        if value is not None:
            return value

    return None





from __future__ import annotations


# Python Modules
import dataclasses
import datetime
import logging

from contextlib import contextmanager
from io import IOBase, StringIO

from pathlib import Path
from typing import Any, ClassVar, Iterable, Mapping, Type, TypeVar

# 3rd Party Modules
import yaml
import yaml_include
from icontract import require

from dateutil.parser import parse as parse_date
from pydantic.dataclasses import dataclass as pydantic_dataclass
from yamlable import YamlAble

# Project Modules
from rwskit.dataclasses_ import DataclassRegistry
from rwskit.types_ import get_qualified_name

log = logging.getLogger(__name__)


T = TypeVar("T", bound="YamlConfig")


# Enable include directives inside yaml files
@contextmanager
def _enable_yaml_include(path: Path):
    tags = ("!inc", "!include")
    loaders = (yaml.Loader, yaml.BaseLoader, yaml.SafeLoader, yaml.FullLoader)
    base_dir = path.parent

    for tag in tags:
        for loader in loaders:
            yaml.add_constructor(tag, yaml_include.Constructor(base_dir=base_dir), loader)

    yield

    for tag in tags:
        for loader in loaders:
            del loader.yaml_constructors[tag]


# According to the python documentation the following (commented) code should
# allow any class that uses the 'YamlConfigMeta' to automatically be viewed
# as a dataclass from the perspective of the type checker. Unless I am doing
# something wrong (very possible), it does not work for me and PyCharm
# does not detect that subclasses should be considered as dataclasses. The
# code functions correctly, but PyCharm will issue warnings about unresolved
# attributes for all subclasses and cannot provide autocompletion.
#
# However, if you annotate each individual class with @dataclass_transform
# PyCharm will recognize the class as a dataclass. This is obviously not ideal
# but better than nothing.

# @dataclass_transform(kw_only_default=True, frozen_default=True)
# class DataclassMeta(type):
#     def __new__(mcs, name, bases, namespace, **kwargs):
#         new_cls = super().__new__(name, bases, namespace, **kwargs)
#         return pydantic_dataclass(new_cls, frozen=True, kw_only=True)
#
#
# class YamlConfigMeta(DataclassMeta, type(YamlAble)):
#     pass


def config_dataclass(cls: Type[T]):
    """A decorator to convert a class to a frozen keyword only pydantic dataclass."""

    return pydantic_dataclass(cls, frozen=True, kw_only=True)


class YamlConfig(YamlAble):
    """A base class for serializable configuration objects.

    Classes that inherit from this class can easily be serialized to and from
    YAML files. Given a YAML file, the class can be reconstructed as long
    as the YAML attributes can be uniquely mapped to a subclass of
    :class:`YamlConfig`.

    Additionally, the configuration can be split across multiple files for
    better modularity using the ``!include`` directive.

    Examples
    --------
    >>> class ChildConfig(YamlConfig):
    ...     id: int
    ...     name: str
    ...     timestamp: datetime.datetime
    >>> class ParentConfig(YamlConfig):
    ...     parent_attr: str = "parent_attr_value"
    ...     child_attr:
    >>> expected_config = ParentConfig(
    ...     id=1,
    ...     child_config=ChildConfig(
    ...         id=2,
    ...         name="child_config",
    ...         timestamp=datetime.datetime.now()
    ... )
    >>> plain_yaml = '''
    ... child_config:
    ...     id: 2
    ...     name: child_config
    ...     timestamp: 2024-11-19 13:55:34.064388
    ...  id: 1
    ... '''
    >>> from_plain_yaml = YamlConfig.loads_yaml(plain_yaml)
    >>> assert from_plain_yaml == expected_config

    The ``!yamlable`` tag can be used  to explicitly tell the YAML parser
    which class to construct. The syntax is ``!yamlable/<fully_qualified_class_name>``.

    >>> tagged_yaml = '''
    ... !yamlable/my_package.my_module.ParentConfig
    ... child_config: !yamlable/my_package.my_module.ChildConfig
    ...     id: 2
    ...     name: child_config
    ...     timestamp: 2024-11-19 13:55:34.064388
    ... id: 1
    ... '''
    >>> assert YamlConfig.loads_yaml(tagged_yaml) == expected_config

    You can use the ``!include`` directive to include other YAML files.
    For example, assume you have the following two YAML files:

    .. code-block:: yaml

        # child_config.yaml

        id: 2
        name: "child_config"
        timestamp: 2024-11-19 13:55:34.064388

    .. code-block:: yaml

        # parent_config.yaml

        id: 1
        child_config: !include child_config.yaml

    You can load the parent config using ``YamlConfig.load_yaml`` as follows:

    >>> YamlConfig.load_yaml("parent_config.yaml")
    """
    # Maintain a registry of all classes that inherit from this base class.
    # Although, I don't plan to use this in a multithreaded context, it is
    # pretty easy to make it thread safe with a lock.
    __registry: ClassVar[DataclassRegistry] = DataclassRegistry()

    default_type_parsers = {
        int: int,
        float: float,
        bool: lambda x: x.lower() in ("true", "1", "t", "y", "yes"),
        str: lambda x: x,
        datetime.datetime: parse_date,
        datetime.date: lambda x: parse_date(x).date(),
    }

    @require(
        lambda self: dataclasses.is_dataclass(self),
        "The class must be a dataclass."
    )
    def __init__(self):
        pass

    def __init_subclass__(cls: Type[YamlConfig], **kwargs: Any):
        """Initialize subclasses to make them suitable configuration objects.

        * Automatically assign the ``__yaml_tag_suffix__`` using the fully
          qualified class name.
        * Convert the class to a ``pydantic.dataclasses.dataclass`` that has
          ``frozen=True`` and ``kw_only=True``.

        Parameters
        ----------
        kwargs : Any
        """
        super().__init_subclass__(**kwargs)

        cls.__yaml_tag_suffix__ = get_qualified_name(cls)

        # Add the subclass to the registry so we can dynamically load the class
        # without having to import it first.
        YamlConfig.__registry.register(cls)  # noqa

    @classmethod
    def get_registered_classes(cls) -> set[Type[YamlConfig]]:
        """Get the set of classes currently registered as configuration objects.

        Returns
        -------
        set[Type[YamlConfig]]
            The set of registered yaml config classes.

        """
        return set(cls.__registry)

    def dumps_plain_yaml(self) -> str:
        """
        Represent the class as plain YAML without any tags.

        .. note::
            It may not be possible to reconstruct the python object from this
            string.

        Returns
        -------
        str
            The object as plain YAML without any tags.
        """

        def _dataclass_to_dict(obj: Any):
            if dataclasses.is_dataclass(obj):
                return {k: _dataclass_to_dict(v) for k, v in dataclasses.asdict(obj).items()}
            elif isinstance(obj, str):
                return obj
            elif isinstance(obj, Mapping):
                return {k: _dataclass_to_dict(v) for k, v in obj.items()}
            elif isinstance(obj, Iterable):
                return [_dataclass_to_dict(v) for v in obj]
            else:
                return obj

        def _transform_sets_to_lists(obj: Any):
            if isinstance(obj, set):
                return [_transform_sets_to_lists(e) for e in obj]
            elif isinstance(obj, Mapping):
                return {k: _transform_sets_to_lists(v) for k, v in obj.items()}
            elif isinstance(obj, str):
                return obj
            elif isinstance(obj, Iterable):
                return [_transform_sets_to_lists(v) for v in obj]
            else:
                return obj

        d = _transform_sets_to_lists(_dataclass_to_dict(self))

        return yaml.safe_dump(d)

    @classmethod
    @require(
        lambda file_path_or_stream: isinstance(file_path_or_stream, (str, Path, IOBase, StringIO)),
        "'file_path_or_stream' must be a string, pathlib.Path, IOBase, or StringIO object."
    )
    def load_yaml(
            cls: Type[T],
            file_path_or_stream: str | Path | IOBase | StringIO,
            safe: bool = True
    ) -> T:
        raw_config = cls._load_raw_yaml(file_path_or_stream, safe)

        # If the deserialized object is an instance of the loader class then
        # we can simply return the object (i.e., it contained all the tags
        # necessary to reconstruct the config).
        if isinstance(raw_config, cls):
            return raw_config

        # Since a config file represents a data class the top level
        # object of a plain yaml config must be a dictionary (i.e., the
        # attributes of the dataclass).
        if not isinstance(raw_config, Mapping):
            raise TypeError(f"Invalid YAML config file.")

        return YamlConfig.__registry.construct_registered_dataclass(raw_config)

    @classmethod
    def _load_raw_yaml(
            cls: Type[T],
            file_path_or_stream: str | Path | IOBase | StringIO,
            safe: bool = True
    ) -> Any:
        # Deserialize the yaml into an object. If the yaml contains known
        # tags it will return the appropriate python classes, otherwise
        # it will return primitive python types (e.g., ints, floats, lists,
        # dicts, etc.)

        yaml_loader = yaml.safe_load if safe else yaml.load

        if isinstance(file_path_or_stream, (str, Path)):
            with open(file_path_or_stream, "rt") as fh:
                # This allows using paths relative to the input config file
                # rather than paths relative to the current working directory.
                with _enable_yaml_include(file_path_or_stream):
                    return yaml_loader(fh)
        else:
            with file_path_or_stream as fh:
                return yaml_loader(fh)



"""Utilities for working with SqlAlchemy."""

from __future__ import annotations

# Python Modules
import logging
from functools import lru_cache
from typing import Any, Literal, Optional, Type, TypeVar, dataclass_transform, get_args

# 3rd Party Modules
import sqlalchemy as sa

from icontract import require
from pydantic.dataclasses import dataclass
from sqlalchemy import orm

# Project Modules
from rwskit.config import YamlConfig, config_dataclass
from rwskit.strings_ import camel_to_snake_case


log = logging.getLogger(__name__)


B = TypeVar("B", bound=orm.DeclarativeBase)
"""
A type extending :class:`~sqlalchemy.orm.DeclarativeBase`.
"""

SqlOperator = Literal[
    "==", "!=", ">", ">=", "<", "<=", "like", "in", "not_in", "is_null", "is_not_null"
]
"""
The supported SQL operators for use in an :class:`SqlBinaryExpression`.
"""


class BaseModel(orm.DeclarativeBase):
    """A base class for creating declarative SqlAlchemy models from.

    This class provides functionality for finding any model derived from
    this clas by their table name.
    """
    __abstract__ = True

    @orm.declared_attr.directive
    def __tablename__(cls):
        return camel_to_snake_case(cls.__name__)

    @classmethod
    @lru_cache
    @require(lambda table_name: table_name.count(".") < 2)
    def find_by_table_name(cls, table_name: str) -> Optional[Type[BaseModel]]:
        """
        Find a model derived from this class by its table name.

        Parameters
        ----------
        table_name : str
            The name of the table whose model class you want to find.

        Returns
        -------
        Type[FindByNameBase], optional
            Returns the model class if the table is found, otherwise ``None``.
        """
        # See: https://stackoverflow.com/a/68862329
        registry = getattr(cls, "registry")

        try:
            find_schema, find_table_name = table_name.split(".", 1)
        except ValueError:
            find_schema, find_table_name = "public", table_name

        for mapper in registry.mappers:
            model = mapper.class_
            table = model.__table__
            candidate_schema = table.schema or "public"
            candidate_table_name = model.__tablename__

            if candidate_schema == find_schema and candidate_table_name == find_table_name:
                return model

        return None


@dataclass(frozen=True, kw_only=False)
class SqlBinaryExpression(YamlConfig):
    """A class that represents the basic binary expression for an SQL column."""

    column: str
    """
    The column name.
    """

    operator: SqlOperator
    """
    The operator to compare the ``column`` and ``value`` with.
    """

    value: Any
    """
    The value used as a comparison.
    """

    @require(
        lambda self: self.operator in get_args(SqlOperator),
        f"The operator must be one of: {get_args(SqlOperator)}",
    )
    def __post_init__(self):
        pass

    def __call__(self, model_or_table: Type[B] | sa.Table) -> sa.BinaryExpression:
        return self.to_expression(model_or_table)

    @require(
        lambda self, model_or_table:
            self.column in model_or_table.columns
            if isinstance(model_or_table, sa.Table) else
            hasattr(model_or_table, self.column),
        "Invalid column"
    )
    @require(
        lambda model_or_table:
            isinstance(model_or_table, sa.Table) or issubclass(model_or_table, orm.DeclarativeBase),
        (
            "The 'model_or_table' must either be an SqlAlchemy Table or an SqlAlchemy ORM model "
            "(subclass of DeclarativeBase)."
        )
    )
    def to_expression(self, model_or_table: Type[B] | sa.Table) -> sa.BinaryExpression:
        """Return a clause that can be used with an SqlAlchemy ``where`` statement.

        Parameters
        ----------
        model_or_table : sqlalchemy.Table
            The table object that contains the column.

        Returns
        -------
        BinaryExpression
            The corresponding SqlAlchemy binary expression.

        """
        column = (
            model_or_table.c[self.column]
            if isinstance(model_or_table, sa.Table) else
            getattr(model_or_table, self.column)
        )
        operator = self.operator

        if operator == "==":
            return column == self.value  # noqa
        if operator == "==" and self.value is None:
            return column.is_(None)
        if operator == "!=":
            return column != self.value  # noqa
        if operator == "!=" and self.value is None:
            return column.isnot(None)
        if operator == '>':
            return column > self.value   # noqa
        if operator == '>=':
            return column >= self.value  # noqa
        if operator == '<':
            return column < self.value   # noqa
        if operator == '<=':
            return column <= self.value  # noqa
        if operator == "like":
            return column.like(self.value)
        if operator == "in":
            return column.in_(self.value)
        if operator == "not_in":
            return column.not_in(self.value)
        if operator == "is_null":
            return column.is_(None)
        if operator == "is_not_null":
            return column.isnot(None)
        else:
            raise ValueError(f'Invalid operator: {operator}')


@dataclass(kw_only=False, frozen=True)
class SqlSelectionCriteria(YamlConfig):
    """A class that represents a conjunction of SqlBinaryExpression."""

    expressions: list[SqlBinaryExpression]
    """
    The list of binary expressions that will be used to filter the query. 
    """

    @require(
        lambda self: all(isinstance(e, SqlBinaryExpression) for e in self.expressions),
        "The expressions must be instances of 'SqlBinaryExpression'.",
    )
    def __post_init__(self):
        pass

    def to_conjunction(self, table: sa.Table) -> sa.BinaryExpression | sa.BooleanClauseList:
        """
        Return a conjunction of binary expressions that can be used with an
        SqlAlchemy ``where`` statement.

        Returns
        -------

        """
        return sa.and_(*[e.to_expression(table) for e in self.expressions])



# Python Modules
import importlib
from types import UnionType

from typing import Any, Type, Union, get_args, get_origin

# 3rd Party Modules
from icontract import require

# Project Modules


def is_union(type_: Any) -> bool:
    """Check if the given ``type_`` is a union of types.

    A type can be a union if it uses the :class:`~typing.Union` type hint or
    the ``|`` operator.

    Parameters
    ----------
    type_ : Any
        The type to check.

    Returns
    -------
    bool
        ``True`` if the given type is a union, otherwise ``False``.

    """
    return get_origin(type_) is Union or isinstance(type_, UnionType)


def is_optional(type_: Any) -> bool:
    """Check if a type is optional.

    An optional type is either defined by class:`typing.Optional` or by
    a :class:`~typing.Union` containing ``None``.

    Parameters
    ----------
    type_ : Any
        The type to check.

    Returns
    -------
    bool
        ``True`` if ``type_`` is optional.

    """
    return is_union(type_) and type(None) in get_args(type_)


@require(
    lambda cls: isinstance(cls, type),
    "The input must be a class type."
)
def get_qualified_name(cls: type) -> str:
    """Get the fully qualified name of this class.

    The fully qualified name is the full path of package names, the module
    name, and class name.

    Parameters
    ----------
    cls : Type[Any]

    Returns
    -------
    str
        The fully qualified name of this class.

    Raises
    ------
    icontract.ViolationError
        If the input ``cls`` is not a type.
    """
    module, name = cls.__module__, cls.__qualname__

    if module is not None and module != "__builtin__":
        name = module + "." + name

    return name


@require(
    lambda qualified_class_name: "." in qualified_class_name,
    "The 'qualified_class_name' must at least refer to a module (the "
    "provided name does not contain a '.')."
)
@require(
    lambda qualified_class_name: bool(qualified_class_name),
    "The 'qualified_class_name' cannot be empty."
)
def type_from_string(qualified_class_name: str) -> type:
    """Import and return the type object from a fully qualified class name.

    Parameters
    ----------
    qualified_class_name : str
        The fully qualified class name.

    Returns
    -------
    type
        The type object corresponding to the fully qualified class name.

    Raises
    ------
    icontract.ViolationError
        If the input ``qualified_class_name`` is ``None`` or empty.
    ModuleNotFoundError
        If the input ``qualified_class_name`` cannot be loaded.
    """
    tokens = qualified_class_name.split('.')
    module_name, class_name = '.'.join(tokens[:-1]), tokens[-1]

    module = importlib.import_module(module_name)

    return getattr(module, class_name)
